#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "gidpost.h"

#ifdef WIN32
#define strcasecmp  _stricmp
#define strdup  _strdup
#else
#include <strings.h>
#endif

/*
  #define DO_FLUSH
  #define EARLY_TERMINATE
*/

#define USE_GP_IN_GROUP

#if defined(USE_GP_IN_GROUP)
#define LPL 3
#define LOC G_elems
#else
#define LPL 1
#define LOC G_nodes
#endif

typedef struct {
  int id;
  float x, y, z;
} SCoord;

typedef struct {
  int id;
  int n1, n2, n3;
} SElem;

#define NUM_NODES   9
#define NUM_ELEMS   8

SCoord G_nodes[ NUM_NODES];
SElem G_elems[ NUM_ELEMS ];

void GenMesh()
{
  SCoord * node;
  SElem * elm1, * elm2;

  int idx;
  idx = 1;
  for ( int i = 0; i < 3; i++ ) {
    for ( int j = 0; j < 3; j++, idx++ ) {
      node = G_nodes + idx -1;
      node->id = idx;
      node->x = (float)(i);
      node->y = (float)(j);
      node->z = 0;
    }
  }
  idx = 0;
  for ( int i = 0; i < 2; i++ ) {
    for ( int j = 0; j < 2; j++, idx+=2 ) {
      elm1 = G_elems + idx;
      elm2 = elm1 + 1;
      elm1->id = idx+1;
      elm1->n1 = i*3 + j + 1;
      elm1->n3 = i*3 + j + 1 + 3 ;
      elm1->n2 = elm1->n3 + 1;
      elm2->id = idx+2;
      elm2->n1 = elm1->n1;
      elm2->n2 = elm1->n1 + 1;
      elm2->n3 = elm1->n2;
    }
  }
}

float Random()
{
  return rand()/(float)(RAND_MAX);
}

void print_help_and_exit( const char *full_prog_name ) {
  // get only the name of the executable
  const char *prog_name = &full_prog_name[ strlen( full_prog_name ) - 1 ];
  for ( ; prog_name > full_prog_name; prog_name-- ) {
    if ( ( *prog_name == '/' ) || ( *prog_name == '\\' ) ) {
      break;
    }
  }
  prog_name++;
  printf( "Usage: %s [ -h] [ -f ascii|bin|hdf5] filename_prefix\n", prog_name );
  printf( "    will create filename_prefix.post.{ msh,res | bin | h5} depending on the choosen format.\n" );
  printf( "    Use: %s -- filename_prefix          if filename_prefix begins with a '-'\n", prog_name );
  exit( 0 );
}

// returns a malloc'ed vector with created dummy results which is to be free'd outside
double *create_dummy_result_block( int num_components, int num_values, int offset_value ) {
  int num_total_values = num_components * num_values;
  double *values = ( double * )malloc( ( size_t)( num_total_values ) * sizeof( double ) );
  if ( values == NULL )
    return NULL;
  // define predictable results
  int offset_values = 0;
  for ( int idx_result = 0; idx_result < num_values; idx_result++ ) {
    for ( int idx_component = 0; idx_component < num_components; idx_component++ ) {
      double my_value = 100 * idx_result + idx_component + 1;
      values[ offset_values ] = my_value + offset_value;
      offset_values++;
    }
  }
  return values;
}

int main( int argc, char *argv[] ) {
#define BUF_SIZE 1024
  char buf[ BUF_SIZE ];
  char test_filename[ BUF_SIZE ];
  char *prefix = NULL;
  char *format = NULL;
  int skip_options = 0;

  for ( int ia = 1; ia < argc; ia++ ) {
    if ( !skip_options && ( argv[ ia ][ 0 ] == '-' ) ) {
      char opt = ( char )tolower( argv[ ia ][ 1 ] );
      if ( ( opt == 'h' ) || !strcasecmp( argv[ ia ], "--help" ) ) {
        print_help_and_exit( argv[ 0 ] );
      } else if ( opt == 'f' ) {
        ia++;
        if ( ia < argc ) {
          format = strdup( argv[ ia ] );
        } else {
          printf( "Missing arguments.\n" );
          print_help_and_exit( argv[ 0 ] );
        }
      } else if ( opt == '-' ) {
        skip_options = 1;
      } else {
        printf( "Unknwon option '%s'.\n", argv[ ia ] );
        print_help_and_exit( argv[ 0 ] );
      }
    } else {
      prefix = strdup( argv[ ia ] );
      break;
    }
  }

  SCoord *node = NULL;
  SElem *elm = NULL;
  int elmi[ 4 ];
  GiD_FILE fdm = 0, fdr = 0;

  GenMesh();

  printf( "version = %s\n", GiD_PostGetVersion() );

  GiD_PostInit();

  // default values
  if ( !prefix )
    prefix = strdup( "test_fd" );
  if ( !format )
    format = strdup( "ascii" );

  int fail = 0;
  if ( !strcasecmp( format, "ascii" ) ) {
    GiD_PostSetFormatReal( "%.8g" );
    snprintf( buf, BUF_SIZE, "%s.post.msh", prefix );
    fdm = GiD_fOpenPostMeshFile( buf, GiD_PostAscii );
    fail = ( fdm == 0 ) ? 1 : 0;
    snprintf( buf, BUF_SIZE, "%s.post.res", prefix );
    // fdr = GiD_fOpenPostResultFile( buf, GiD_PostAsciiZipped );
    fdr = GiD_fOpenPostResultFile( buf, GiD_PostAscii );
    fail += ( fdr == 0 ) ? 1 : 0;
    printf( "Creating ASCII gid post files '%s.post.msh' and '%s.post.res'.\n", prefix, prefix );
  } else if ( !strcasecmp( format, "bin" ) ) {
    snprintf( buf, BUF_SIZE, "%s.post.bin", prefix );
    fdm = fdr = GiD_fOpenPostResultFile( buf, GiD_PostBinary );
    fail = ( fdm == 0 ) ? 1 : 0;
    printf( "Creating BINary gid post file '%s'.\n", buf );
  } else if ( !strcasecmp( format, "hdf5" ) ) {
    snprintf( buf, BUF_SIZE, "%s.post.h5", prefix );
    printf( "Creating HDF5 gid post file '%s'.\n", buf );
    fdm = fdr = GiD_fOpenPostResultFile( buf, GiD_PostHDF5 );
    fail = ( fdm == 0 ) ? 1 : 0;
  } else {
    printf( "Unkown format '%s'.\n", format );
    print_help_and_exit( argv[ 0 ] );
  }
  strcpy( test_filename, buf );

  if ( fail ) {
    printf( "Error opening file with format '%s'.\n", format );
    exit( 1 );
  }

  // // free strdup memory
  // free( format );
  // format = NULL;
  // free( prefix );
  // prefix = NULL;

  /* write mesh info */
  GiD_fBeginMeshColor( fdm, "TestMsh", GiD_2D, GiD_Triangle, 3, 0, 0.99, 0 );
  /* coordinates */
  const int write_array = 1;
  if ( write_array ) {
    int ids[ NUM_NODES ];
    double xyz[ NUM_NODES * 3 ];
    for ( int i = 0; i < NUM_NODES; i++ ) {
      ids[ i ] = G_nodes[ i ].id;
      xyz[ i * 3 + 0 ] = G_nodes[ i ].x;
      xyz[ i * 3 + 1 ] = G_nodes[ i ].y;
      xyz[ i * 3 + 2 ] = G_nodes[ i ].z;
    }
    GiD_fWriteCoordinatesIdBlock( fdm, NUM_NODES, ids, xyz );
    // if id's are = 1..NUM_NODES
    // GiD_fWriteCoordinatesBlock( fdm, NUM_NODES, xyz );
  } else {
    GiD_fBeginCoordinates( fdm );
    for ( int i = 0; i < NUM_NODES; i++ ) {
      node = G_nodes + i;
      GiD_fWriteCoordinates( fdm, node->id, node->x, node->y, node->z );
    }
    GiD_fEndCoordinates( fdm );
  }
  /* elements */
  if ( write_array ) {
    int ids[ NUM_ELEMS ];
    int connectivities[ NUM_ELEMS * 3 ];
    int mat_ids[ NUM_ELEMS ];
    for ( int i = 0; i < NUM_ELEMS; i++ ) {
      ids[ i ] = G_elems[ i ].id;
      connectivities[ i * 3 + 0 ] = G_elems[ i ].n1;
      connectivities[ i * 3 + 1 ] = G_elems[ i ].n2;
      connectivities[ i * 3 + 2 ] = G_elems[ i ].n3;
      mat_ids[ i ] = 2;
    }
    GiD_fWriteElementsIdMatBlock( fdm, NUM_ELEMS, ids, connectivities, mat_ids );
    // if id's are = 1..NUM_ELEMS then we could have used
    // GiD_fWriteElementsMatBlock( fdm, NUM_ELEMS, connectivities, mat_ids);
    // if id's no consecutives, but no material_ids then
    // GiD_fWriteElementsIdBlock( fdm, NUM_ELEMS ids, connectivities);
    // if id's are = 1..NUM_ELEMS and no material_ids them
    // GiD_fWriteElementsBlock( fdm, NUM_ELEMS, connectivities);
  } else {
    GiD_fBeginElements( fdm );
    for ( int i = 0; i < NUM_ELEMS; i++ ) {
      elm = G_elems + i;
      elmi[ 0 ] = elm->n1;
      elmi[ 1 ] = elm->n2;
      elmi[ 2 ] = elm->n3;
      elmi[ 3 ] = 2;
      GiD_fWriteElementMat( fdm, elm->id, elmi );
    }
    GiD_fEndElements( fdm );
  }
  GiD_fEndMesh( fdm );

  /* now results info */


// #define SIMPLE_RESULT

#ifdef SIMPLE_RESULT
  if ( write_array ) {
    const int num_components = 1;
    const int offset_value = 0;
    double *scalar = create_dummy_result_block( num_components, NUM_NODES, offset_value );
    // GiD_fWriteResultBlock includes BeginResult()/BeginValues() and EndValues()/EndResult()
    const char *unit_name = "m"; // just an example
    GiD_fWriteResultBlock( fdr, "EscalarNodos", "Analysis", 1.0, GiD_Scalar, GiD_OnNodes, NULL, NULL, 0, NULL, unit_name,
                           NUM_NODES, NULL, num_components, scalar);
    free( scalar );
  } else {
    GiD_fBeginResult( fdr, "EscalarNodos", "Analysis", 1.0, GiD_Scalar, GiD_OnNodes, NULL, NULL, 0, NULL );
    for ( int i = 0; i < NUM_NODES; i++ ) {
      GiD_fWriteScalar( fdr, G_nodes[ i ].id, Random() );
    }
    GiD_fEndResult( fdr );
  }
#else // SIMPLE_RESULT
  /* write the gauss points */
  GiD_fBeginGaussPoint(fdr, "element_gp", GiD_Triangle, NULL, 3, 0, 1);  
  GiD_fEndGaussPoint(fdr);
  
  /* scalar result over G_nodes */
  if ( write_array ) {
    printf( "no support for Result Groups when writing ResultBlocks.\n" );
  }
#if defined(USE_GP_IN_GROUP)
  GiD_fBeginResultGroup(fdr, "Analysis", 1.0, GiD_OnGaussPoints,"element_gp" );
#else
  GiD_fBeginResultGroup(fdr, "Analysis", 1.0, GiD_OnNodes, NULL);
#endif
  // 'path//to//result' is a GiD convention to group results. it can also be used with GiD_fWriteResultBlock and GiD_fBeginResultHeader
  GiD_fResultDescription(fdr, "gaussian//EscalarNodosG", GiD_Scalar);
  GiD_fResultUnit( fdr, "mm" );
  GiD_fResultDescriptionDim(fdr, "gaussian//VectorNodosG", GiD_Vector, 4);
  GiD_fResultUnit( fdr, "unit" );
  GiD_fResultDescription(fdr, "gaussian//MatrixG", GiD_Matrix);
  GiD_fResultUnit( fdr, "N�m" );
  GiD_fResultDescription(fdr, "gaussian//Local AxesG", GiD_LocalAxes);
  for ( int i = 0; i < sizeof(LOC)/sizeof(LOC[0]); i++ ) {
    for ( int j = 0; j < LPL; j++) {
      GiD_fWriteScalar(fdr, LOC[i].id, Random());
      GiD_fWriteVectorModule(fdr, LOC[i].id, Random(), Random(), Random(),-1);
      GiD_fWrite3DMatrix(fdr, LOC[i].id, Random(), Random(), Random(),
                         Random(), Random(), Random());
      GiD_fWriteLocalAxes(fdr, LOC[i].id, Random(), Random(), Random());
    }
  }
  GiD_fEndResult(fdr);
  
  /* #else */

  /* scalar results over gauss points: Triangle + 3 GP */

  if ( write_array ) {
    const int num_components = 1;
    const int offset_value = 1000;
    const int num_gp_per_element = 3;
    double *scalar = create_dummy_result_block( num_components, num_gp_per_element * NUM_ELEMS, offset_value );
    // GiD_fWriteResultBlock includes BeginResult()/BeginValues() and EndValues()/EndResult()
    int list_ids[ NUM_ELEMS * 3 ];
    int idx_id = 0;
    for ( int i = 0; i < NUM_ELEMS; i++ ) {
      // all gauss points inside the same element, share the same element id
      for ( int igp = 0; igp < num_gp_per_element; igp++ ) {
        list_ids[ idx_id ] = G_elems[ i ].id;
        idx_id++;
      }
    }

    const char *unit_name = "m"; // just an example
    GiD_fWriteResultBlock( fdr, "SobreGPTriangle3", "Analysis", 1.0, GiD_Scalar, GiD_OnGaussPoints, "element_gp", NULL,
                           0, NULL, unit_name,
                           num_gp_per_element * NUM_ELEMS, list_ids, num_components, scalar );
    free( scalar );
  } else {
    GiD_fBeginResult( fdr, "SobreGPTriangle3", "Analysis", 1.0, GiD_Scalar, GiD_OnGaussPoints, "element_gp", NULL, 0,
                      NULL );
    for ( int i = 0; i < NUM_ELEMS; i++ ) {
      for ( int j = 0; j < 3; j++ )
        GiD_fWriteScalar( fdr, i + 1, Random() );
    }
    GiD_fEndResult( fdr );
  }

  if ( write_array ) {
    const int num_components = 1;
    const int offset_value = 2000;
    double *scalar = create_dummy_result_block( num_components, NUM_NODES, offset_value );
    // GiD_fWriteResultBlock includes BeginResult()/BeginValues() and EndValues()/EndResult()
    int list_ids[ NUM_NODES ];
    for ( int i = 0; i < NUM_NODES; i++ ) {
      list_ids[ i ] = G_nodes[ i ].id;
    }
    const char *unit_name = "m"; // just an example
    GiD_fWriteResultBlock( fdr, "EscalarNodos", "Analysis", 1.0, GiD_Scalar, GiD_OnNodes, NULL, NULL, 0, NULL, unit_name,
                           NUM_NODES, list_ids, num_components, scalar );
    free( scalar );
  } else {
    GiD_fBeginResult( fdr, "EscalarNodos", "Analysis", 1.0, GiD_Scalar, GiD_OnNodes, NULL, NULL, 0, NULL );
    for ( int i = 0; i < NUM_NODES; i++ ) {
      GiD_fWriteScalar( fdr, G_nodes[ i ].id, Random() );
    }
    GiD_fEndResult( fdr );
  }

  /* vector result over G_nodes */
  if ( write_array ) {
    const int num_components = 3; // xyz vector
    const int offset_value = 3000;
    double *scalar = create_dummy_result_block( num_components, NUM_NODES, offset_value );
    // GiD_fWriteResultBlock includes BeginResult()/BeginValues() and EndValues()/EndResult()
    int list_ids[ NUM_NODES ];
    for ( int i = 0; i < NUM_NODES; i++ ) {
      list_ids[ i ] = G_nodes[ i ].id;
    }
    const char *unit_name = "m/s"; // just an example
    GiD_fWriteResultBlock( fdr, "VectorNodos", "Analysis", 1.0, GiD_Vector, GiD_OnNodes, NULL, NULL, 0, NULL, unit_name,
                           NUM_NODES, list_ids, num_components, scalar );
    free( scalar );
  } else {
    /*
    GiD_fBeginResult("VectorNodos", "Analysis", 1.0, GiD_Vector, GiD_OnNodes,
    NULL, NULL, 0, NULL);*/
    GiD_fBeginResultHeader( fdr, "VectorNodos", "Analysis", 1.0, GiD_Vector, GiD_OnNodes, NULL );
    for ( int i = 0; i < NUM_NODES; i++ ) {
      GiD_fWriteVector( fdr, G_nodes[ i ].id, Random(), Random(), Random() );
    }
    GiD_fEndResult( fdr );
  }

#ifndef DISABLE_CODE

#ifdef DO_FLUSH
  GiD_fFlushPostFile(fdr);
#endif

  /* matrix result */
  if ( write_array ) {
    const int num_components = 6; // 3d matrix
    const int offset_value = 4000;
    double *scalar = create_dummy_result_block( num_components, NUM_NODES, offset_value );
    // GiD_fWriteResultBlock includes BeginResult()/BeginValues() and EndValues()/EndResult()
    const char *unit_name = "Kg/ms^2"; // just an example
    GiD_fWriteResultBlock( fdr, "Matrix", "Analysis", 1.0, GiD_Matrix, GiD_OnNodes, NULL, NULL, 0, NULL, unit_name,
                           NUM_NODES, NULL, num_components, scalar );
    free( scalar );
  } else {
    GiD_fBeginResult( fdr, "Matrix", "Analysis", 1.0, GiD_Matrix, GiD_OnNodes, NULL, NULL, 0, NULL );
    for ( int i = 0; i < NUM_NODES; i++ ) {
      GiD_fWrite3DMatrix( fdr, i + 1, Random(), Random(), Random(), Random(), Random(), Random() );
    }
    GiD_fEndResult( fdr );
  }

#ifdef EARLY_TERMINATE
  abort();
#endif

  /* local axes result */
  if ( write_array ) {
    const int num_components = 3; // local axis
    const int offset_value = 5000;
    double *scalar = create_dummy_result_block( num_components, NUM_NODES, offset_value );
    // GiD_fWriteResultBlock includes BeginResult()/BeginValues() and EndValues()/EndResult()
    const char *unit_name = NULL; // just an example
    GiD_fWriteResultBlock( fdr, "Local Axes", "Analysis", 1.0, GiD_LocalAxes, GiD_OnNodes, NULL, NULL, 0, NULL, unit_name,
                           NUM_NODES, NULL, num_components, scalar );
    free( scalar );
  } else {
    GiD_fBeginResult( fdr, "Local Axes", "Analysis", 1.0, GiD_LocalAxes, GiD_OnNodes, NULL, NULL, 0, NULL );
    for ( int i = 0; i < NUM_NODES; i++ ) {
      GiD_fWriteLocalAxes( fdr, i + 1, Random(), Random(), Random() );
    }
    GiD_fEndResult( fdr );
  }
#endif /* DISABLE_CODE */

#endif /* SIMPLE_RESULT */
  
  if ( fdm && ( fdm != fdr ) ) {
    // if fdm != fdr, then we are in ascii format, we have 2 separated files: one for mesh and another for results
    GiD_fClosePostMeshFile( fdm );
  }
  GiD_fClosePostResultFile(fdr);

  GiD_PostDone();

  printf( "%s written with format '%s'", test_filename, format );
  if ( write_array ) {
    printf( " using *Block() API" );
  }
  printf( "\n" );

  // free strdup memory
  free( format );
  format = NULL;
  free( prefix );
  prefix = NULL;

  return 0;
}
